"""Base Model class that implements CRUD methods for database entities based on Pydantic dataclasses"""
import asyncio
from datetime import datetime
from enum import Enum, EnumMeta
import inspect
from typing import Any, Callable, Dict, Generator, List, Optional, Sequence, Set, Union
from uuid import UUID as stdlib_uuid

from aenum import Enum as AEnum, EnumMeta as AEnumMeta
from pydantic import BaseModel, PositiveFloat, PositiveInt
from pydantic.types import UUID4
from sqlalchemy import (
    and_,
    Boolean,
    Column,
    Enum as SQLEnum,
    Float,
    Integer,
    MetaData,
    Table,
    TIMESTAMP,
    VARCHAR,
)
from sqlalchemy.dialects.postgresql import insert, JSONB, UUID as sqlalchemy_uuid
from sqlalchemy.schema import FetchedValue
from sqlalchemy.sql.base import ImmutableColumnCollection
from sqlalchemy.sql.elements import BinaryExpression, UnaryExpression

from pynocular.engines import DBEngine, DBInfo
from pynocular.exceptions import (
    DatabaseModelMisconfigured,
    DatabaseModelMissingField,
    DatabaseRecordNotFound,
    InvalidFieldValue,
    InvalidMethodParameterization,
    InvalidTextRepresentation,
    NestedDatabaseModelNotResolved,
)
from pynocular.nested_database_model import NestedDatabaseModel


def is_valid_uuid(string: str) -> bool:
    """Check if a string is a valid UUID

    Args:
        string: the string to check

    Returns:
        Whether or not the string is a well-formed UUIDv4

    """
    try:
        stdlib_uuid(string, version=4)
        return True
    except (TypeError, AttributeError, ValueError):
        return False


class UUID_STR(str):
    """A string that represents a UUID4 value"""

    @classmethod
    def __get_validators__(cls) -> Generator:
        """Get the validators for the given class"""
        yield cls.validate

    @classmethod
    def validate(cls, v: Any) -> str:
        """Function to validate the value

        Args:
            v: The value to validate

        """
        if isinstance(v, stdlib_uuid) or (isinstance(v, str) and is_valid_uuid(v)):
            return str(v)
        else:
            raise ValueError("invalid UUID string")


def nested_model(
    db_model_class: "DatabaseModel", reference_field: str = None
) -> Callable:
    """Generate a NestedModel class with dynamic model references

    Args:
        db_model_class: The specific model class that will be nested. This will be a
            subclass of `DatabaseModel`
        reference_field: The name of the field on the database table that this nested
            model references.

    """

    class NestedModel:
        """NestedModel type for NestedDatabaseModels"""

        reference_field_name = reference_field

        @classmethod
        def __get_validators__(cls) -> Generator:
            """Get the validators for the given class"""
            yield cls.validate

        @classmethod
        def validate(cls, v: Union[UUID_STR, "DatabaseModel"]) -> NestedDatabaseModel:
            """Validate value and generate a nested database model"""
            # If value is a uuid then create a NestedDatabaseModel, otherwise just
            # Set the DatabaseModel as the value
            if is_valid_uuid(v):
                return NestedDatabaseModel(db_model_class, v)
            else:
                return NestedDatabaseModel(db_model_class, v.get_primary_id(), v)

    return NestedModel


def database_model(table_name: str, database_info: DBInfo) -> "DatabaseModel":
    """Decorator that adds SQL functionality to Pydantic BaseModel objects

    Args:
        table_name: Name of the table this model represents in the database
        database_info: Database connection info for the database to connect to

    Raises:
        DatabaseModelMisconfigured: Raised when class with this decorator is not a pydantic.BaseModel
            subclass. We depend on the class implementing a some specific things and currently don't
            support any other type of dataclass.

    """

    def wrapped(cls):
        if BaseModel not in inspect.getmro(cls):
            raise DatabaseModelMisconfigured(
                "Model is not subclass of pydantic.BaseModel"
            )

        cls.__bases__ += (DatabaseModel,)
        cls.initialize_table(table_name, database_info)

        return cls

    return wrapped


class DatabaseModel:
    """Adds database functionality to a Pydantic BaseModel

    A DatabaseModel is a Pydantic based model along with a SQLAlchemy
    table object. This allows us to use the same object for both
    database queries and HTTP requests

    """

    # Define metadata for the database connection on the class level so we don't
    # have to recaluclate the table for each database call
    _table: Table = None
    _database_info: DBInfo = None

    # We may have times where we need a compound primary key.
    # We store each one into this list and have our query functions
    # handle using it
    _primary_keys: List[Column] = None

    # Some fields are exclusively produced by the database server
    # For all save operations, we need to get those values from the database
    # These are the server_default and server_onupdate functions in SQLAlchemy
    _db_managed_fields: List[str] = None

    # The following tables track which attributes on the model are nested model
    # references
    # Some nested model attributes may have different names than their actual db table;
    # For example; on an App we may have an `org` attribute but the db field is
    # `organzation_id`

    # In order to manage this we also need maps from attribute name to table_field_name
    # and back
    _nested_model_attributes: Set[str] = None
    _nested_attr_table_field_map: Dict[str, str] = None
    _nested_table_field_attr_map: Dict[str, str] = None

    # This can be used to access the table when defining where expressions
    columns: ImmutableColumnCollection = None

    @classmethod
    def initialize_table(cls, table_name: str, database_info: DBInfo) -> None:
        """Returns a SQLAlchemy table definition to expose SQLAlchemy functions

        This method should cache the Table on the __table__ class property.
        We don't want to have to recaluclate the table for every SQL call,
        so it's desirable to cache this at the class level.

        Returns:
            A Table object based on the Field properties defined from the Pydantic model

        Raises:
            DatabaseModelMisconfigured: When the class does not defined certain properties;
                or cannot be converted to a Table

        """
        cls._primary_keys = []
        cls._database_info = database_info
        cls._db_managed_fields = []
        cls._nested_attr_table_field_map = {}
        cls._nested_table_field_attr_map = {}
        cls._nested_model_attributes = set()

        columns = []
        for field in cls.__fields__.values():
            name = field.name
            is_nullable = not field.required
            is_primary_key = field.field_info.extra.get("primary_key", False)
            fetch_on_create = field.field_info.extra.get("fetch_on_create", False)
            fetch_on_update = field.field_info.extra.get("fetch_on_update", False)

            if field.type_ is str:
                type = VARCHAR
            elif field.type_.__name__ == "ConstrainedStrValue":
                # This is because pydantic is doing some kind of dynamic type construction.
                # See: https://github.com/samuelcolvin/pydantic/blob/e985857e5a9ede8d346b010a5a039aa84a089826/pydantic/types.py#L245-L263
                length = field.field_info.max_length
                type = VARCHAR(length)
            elif (
                field.type_ in (int, PositiveInt)
                or field.type_.__name__ == "ConstrainedIntValue"
            ):
                type = Integer
            elif (
                field.type_ in (float, PositiveFloat)
                or field.type_.__name__ == "ConstrainedFloatValue"
            ):
                type = Float
            elif field.type_.__class__ in (AEnumMeta, EnumMeta):
                type = SQLEnum(field.type_)
            elif field.type_ is bool:
                type = Boolean
            elif field.type_ in (dict, Dict):
                type = JSONB(none_as_null=True)
            elif field.type_ in (UUID4, stdlib_uuid, UUID_STR):
                type = sqlalchemy_uuid()
            elif field.type_ is datetime:
                type = TIMESTAMP(timezone=True)
            elif field.type_.__name__ == "NestedModel":
                cls._nested_model_attributes.add(name)
                # If the field name on the NestedModel type is not None, use that for the
                # column name
                if field.type_.reference_field_name is not None:
                    cls._nested_attr_table_field_map[
                        name
                    ] = field.type_.reference_field_name
                    cls._nested_table_field_attr_map[
                        field.type_.reference_field_name
                    ] = name
                    name = field.type_.reference_field_name

                # Assume all IDs are UUIDs for now
                type = sqlalchemy_uuid()
            # TODO - how are people using this today? Is there a class we need to make or can we reuse one
            # elif field.type_ is bit:
            #     type = Bit
            else:
                raise DatabaseModelMisconfigured(f"Unsupported type {field.type_}")

            column = Column(
                name, type, primary_key=is_primary_key, nullable=is_nullable
            )

            if fetch_on_create:
                column.server_default = FetchedValue()
                cls._db_managed_fields.append(name)

            if fetch_on_update:
                column.server_onupdate = FetchedValue()
                cls._db_managed_fields.append(name)

            if is_primary_key:
                cls._primary_keys.append(column)

            columns.append(column)

        cls._table = Table(table_name, MetaData(), *columns)
        cls.columns = cls._table.c

    @classmethod
    async def get_with_refs(cls, *args: Any, **kwargs: Any) -> "DatabaseModel":
        """Gets the DatabaseModel associated with any nested key references resolved

        Args:
            args: The column id for the object's primary key
            kwargs: The columns and ids that make up the object's composite primary key

        Returns:
            A DatabaseModel object representing the record in the db if one exists

        """
        obj = await cls.get(*args, **kwargs)
        gatherables = [
            (getattr(obj, prop_name)).fetch()
            for prop_name in cls._nested_model_attributes
        ]
        await asyncio.gather(*gatherables)

        return obj

    @classmethod
    async def get(cls, *args: Any, **kwargs: Any) -> "DatabaseModel":
        """Gets the DatabaseModel for the given primary key value(s)

        Args:
            args: The column id for the object's primary key
            kwargs: The columns and ids that make up the object's composite primary key

        Returns:
            A DatabaseModel object representing the record in the db if one exists

        Raises:
            InvalidMethodParameterization: An invalid parameter configuration was passed in.
                This method should only receive one arg or >= one kwargs. Any other
                combination of parameters is invalid.

        """
        if (
            (len(args) > 1)
            or (len(args) == 1 and len(kwargs) > 0)
            or (len(args) == 1 and len(cls._primary_keys) > 1)
            or (len(args) == 0 and len(kwargs) == 0)
        ):
            raise InvalidMethodParameterization("get", args=args, kwargs=kwargs)

        if len(args) == 1:
            primary_key_dict = {cls._primary_keys[0].name: args[0]}
        else:
            primary_key_dict = kwargs

        original_primary_key_dict = primary_key_dict.copy()
        where_expressions = []
        for primary_key in cls._primary_keys:
            primary_key_value = primary_key_dict.pop(primary_key.name)
            where_expressions.append(primary_key == primary_key_value)

        records = await cls.select(where_expressions=where_expressions, limit=1)
        if len(records) == 0:
            raise DatabaseRecordNotFound(cls._table.name, **original_primary_key_dict)

        return records[0]

    @classmethod
    async def get_list(cls, **kwargs: Any) -> List["DatabaseModel"]:
        """Fetches the DatabaseModel for based on the provided kwargs

        Args:
            kwargs: The filterable key/value pairs for the where clause. These will be
                `and`ed together

        Returns:
            List of DatabaseModel objects

        Raises:
            DatabaseModelMisconfigured: The class is missing a database table
            DatabaseModelMissingField: One of the fields provided in the query does not
                exist on the database table

        """
        where_clause_list = []
        for field_name, db_field_value in kwargs.items():
            db_field_name = cls._nested_attr_table_field_map.get(field_name, field_name)

            try:
                db_field = getattr(cls._table.c, db_field_name)
            except AttributeError:
                raise DatabaseModelMissingField(cls.__name__, db_field_name)

            if isinstance(db_field_value, list):
                exp = db_field.in_(db_field_value)
            else:
                exp = db_field == db_field_value

            where_clause_list.append(exp)

        return await cls.select(where_expressions=where_clause_list)

    @classmethod
    async def select(
        cls,
        where_expressions: Optional[List[BinaryExpression]] = None,
        order_by: Optional[List[UnaryExpression]] = None,
        limit: Optional[int] = None,
    ) -> List["DatabaseModel"]:
        """Execute a SELECT on the DatabaseModel table with the given parameters

        Args:
            where_expressions: A list of BinaryExpressions for the table that will be
                `and`ed together for the where clause of the SELECT
            order_by: A list of criteria for the order_by clause
            limit: The number of instances to return

        Returns:
            A list of DatabaseModel instances

        Raises:
            DatabaseModelMisconfigured: The class is missing a database table

        """
        async with (
            await DBEngine.transaction(cls._database_info, is_conditional=True)
        ) as conn:
            query = cls._table.select()
            if where_expressions is not None and len(where_expressions) > 0:
                query = query.where(and_(*where_expressions))
            if order_by is not None and len(order_by) > 0:
                query = query.order_by(*order_by)
            if limit is not None and limit > 0:
                query = query.limit(limit)

            try:
                result = await conn.execute(query)
            # The value was the wrong type. This usually happens with UUIDs.
            except InvalidTextRepresentation as e:
                raise InvalidFieldValue(message=e.diag.message_primary)
            records = await result.fetchall()

            return [cls.from_dict(dict(record)) for record in records]

    @classmethod
    async def create(cls, **data) -> "DatabaseModel":
        """Create a new instance of the this DatabaseModel and save it

        Args:
            kwargs: The parameters for the instance

        Returns:
            The new DatabaseModel instance

        """
        new = cls(**data)
        await new.save()

        return new

    @classmethod
    async def create_list(cls, models: List["DatabaseModel"]) -> List["DatabaseModel"]:
        """Create new batch of records in one query

        This will mutate the provided models to include db managed column values.

        Args:
            models: List of database models to persist

        Returns:
            list of new database models that have been saved

        """
        if not models:
            return []

        values = []
        for model in models:
            dict_obj = model.to_dict()
            for field in cls._db_managed_fields:
                # Remove any fields that the database calculates
                del dict_obj[field]
            values.append(dict_obj)

        async with (
            await DBEngine.transaction(cls._database_info, is_conditional=False)
        ) as conn:
            result = await conn.execute(
                insert(cls._table).values(values).returning(cls._table)
            )
            # Set db managed column information on the object
            rows = await result.fetchall()
            for row, model in zip(rows, models):
                record_dict = dict(row)
                for column in cls._db_managed_fields:
                    col_val = record_dict.get(column)
                    if col_val is not None:
                        setattr(model, column, col_val)

        return models

    @classmethod
    async def delete_records(cls, **kwargs: Any) -> None:
        """Execute a DELETE on a DatabaseModel with the provided kwargs

        Args:
            kwargs: The filterable key/value pairs for the where clause. These will be
                `and`ed together

        Raises:
            DatabaseModelMisconfigured: The class is missing a database table
            DatabaseModelMissingField: One of the fields provided in the query does not
                exist on the database table

        """
        where_clause_list = []
        for field_name, db_field_value in kwargs.items():
            db_field_name = cls._nested_attr_table_field_map.get(field_name, field_name)

            try:
                db_field = getattr(cls._table.c, db_field_name)
            except AttributeError:
                raise DatabaseModelMissingField(cls.__name__, db_field_name)

            if isinstance(db_field_value, list):
                exp = db_field.in_(db_field_value)
            else:
                exp = db_field == db_field_value

            where_clause_list.append(exp)

        async with (
            await DBEngine.transaction(cls._database_info, is_conditional=False)
        ) as conn:
            query = cls._table.delete().where(and_(*where_clause_list))
            try:
                await conn.execute(query)
            # The value was the wrong type. This usually happens with UUIDs.
            except InvalidTextRepresentation as e:
                raise InvalidFieldValue(message=e.diag.message_primary)

    @classmethod
    async def update_record(cls, **kwargs: Any) -> "DatabaseModel":
        """Update a record associated with this DatabaseModel

        Notes:
            the primary key must be in the kwargs

        Args:
            kwargs: The values to update.

        Returns:
            The updated DatabaseModel

        """
        where_expressions = []
        primary_key_dict = {}
        for primary_key in cls._primary_keys:
            primary_key_value = kwargs.pop(primary_key.name)
            where_expressions.append(primary_key == primary_key_value)
            primary_key_dict[primary_key.name] = primary_key_value

        modified_kwargs = {}
        for field_name, value in kwargs.items():
            db_field_name = cls._nested_attr_table_field_map.get(field_name, field_name)
            modified_kwargs[db_field_name] = value

        updated_records = await cls.update(where_expressions, modified_kwargs)
        if len(updated_records) == 0:
            raise DatabaseRecordNotFound(cls._table.name, **primary_key_dict)
        return updated_records[0]

    @classmethod
    async def update(
        cls, where_expressions: Optional[List[BinaryExpression]], values: Dict[str, Any]
    ) -> List["DatabaseModel"]:
        """Execute an UPDATE on a DatabaseModel table with the given parameters

        Args:
            where_expressions: A list of BinaryExpressions for the table that will be
                `and`ed together for the where clause of the UPDATE
            values: The field and values to update all records to that match the
                where_expressions

        Returns:
            The updated DatabaseModels

        Raises:
            DatabaseModelMisconfigured: The class is missing a database table

        """
        async with (
            await DBEngine.transaction(cls._database_info, is_conditional=False)
        ) as conn:
            query = (
                cls._table.update()
                .where(and_(*where_expressions))
                .values(**values)
                .returning(cls._table)
            )
            try:
                results = await conn.execute(query)
            # The value was the wrong type. This usually happens with UUIDs.
            except InvalidTextRepresentation as e:
                raise InvalidFieldValue(message=e.diag.message_primary)

            return [cls.from_dict(dict(record)) for record in await results.fetchall()]

    async def save(self, include_nested_models=False) -> None:
        """Update the database record this object represents with its current state

        Args:
            include_nested_models: If True, any nested models should get saved before
                this object gets saved

        """

        dict_self = self.to_dict()

        primary_key_names = [primary_key.name for primary_key in self._primary_keys]

        for field in self._db_managed_fields:
            if field in primary_key_names and dict_self[field] is not None:
                continue

            # Remove any fields that the database calculates
            del dict_self[field]

        async with (
            await DBEngine.transaction(self._database_info, is_conditional=False)
        ) as conn:
            # If flag is set, first try to persist any nested models. This needs to
            # happen inside of the transaction so if something fails everything gets
            # rolled back
            if include_nested_models:
                for attr_name in self._nested_model_attributes:
                    try:
                        obj = getattr(self, attr_name)
                        if obj is not None:
                            await obj.save()
                    except NestedDatabaseModelNotResolved:
                        # If the object was never resolved than it already exists in the
                        # DB and the DB has the latest state
                        continue

            record = await conn.execute(
                insert(self._table)
                .values(dict_self)
                .on_conflict_do_update(index_elements=primary_key_names, set_=dict_self)
                .returning(self._table)
            )

            row = await record.fetchone()

            for field in self._db_managed_fields:
                setattr(self, field, row[field])

    def get_primary_id(self) -> Any:
        """Standard interface for returning the id of a field

        This assumes that there is a single primary id, otherwise this returns `None`

        Returns:
            The ID value for this DatabaseModel instance

        """
        if len(self._primary_keys) > 1:
            return None

        return getattr(self, self._primary_keys[0].name)

    async def fetch(self, resolve_references: bool = False) -> None:
        """Gets the latest of the object from the database and updates itself

        Args:
            resolve_references: If True, resolve any nested key references

        """
        # Get the latest version of self
        get_params = {
            primary_key.name: getattr(self, primary_key.name)
            for primary_key in self._primary_keys
        }
        if resolve_references:
            new_self = await self.get_with_refs(**get_params)
        else:
            new_self = await self.get(**get_params)

        for attr_name, new_attr_val in new_self.dict().items():
            setattr(self, attr_name, new_attr_val)

    async def delete(self) -> None:
        """Delete this record from the database"""

        async with (
            await DBEngine.transaction(self._database_info, is_conditional=False)
        ) as conn:
            where_expressions = [
                getattr(self._table.c, pkey.name) == getattr(self, pkey.name)
                for pkey in self._primary_keys
            ]
            query = self._table.delete().where(and_(*where_expressions))
            try:
                await conn.execute(query)
            # The value was the wrong type. This usually happens with UUIDs.
            except InvalidTextRepresentation as e:
                raise InvalidFieldValue(message=e.diag.message_primary)

    @classmethod
    def from_dict(cls, _dict: Dict[str, Any]) -> "DatabaseModel":
        """Instantiate a DatabaseModel object from a dict record

        Note:
            This is the base implementation and is set up so classes that subclass this
            one don't have to make this boilerplate if they don't need to

        Args:
            _dict: The dictionary form of the DatabaseModel

        Returns:
            The DatabaseModel object

        """
        modified_dict = {}
        for key, value in _dict.items():
            modified_key = cls._nested_table_field_attr_map.get(key, key)
            modified_dict[modified_key] = value
        return cls(**modified_dict)

    def to_dict(
        self, serialize: bool = False, include_keys: Optional[Sequence] = None
    ) -> Dict[str, Any]:
        """Create a dict from the DatabaseModel object

        Note:
            This implementation is only valid if __base_props__ is set for the instance

        Args:
            serialize: A flag determining whether or not to serialize enum types into
                strings
            include_keys: Set of keys that should be included in the results. If not
                provided or empty, all keys will be included.

        Returns:
            A dict of the DatabaseObject object

        Raises:
            NotImplementedError: This function implementation is being used without
                __base_props__ being set

        """
        _dict = {}
        for prop_name, prop_value in self.dict().items():
            if serialize:
                if isinstance(prop_value, Enum):
                    prop_value = prop_value.name
                elif isinstance(prop_value, AEnum):
                    prop_value = prop_value.value

            if prop_name in self._nested_model_attributes:
                # self.dict() will serialize any BaseModels into a dict so fetch the
                # actual object from self
                temp_prop_value = getattr(self, prop_name)
                prop_name = self._nested_attr_table_field_map.get(prop_name, prop_name)
                # temp_prop_value can be `None` if the nested key is optional
                if temp_prop_value is not None:
                    prop_value = temp_prop_value.get_primary_id()

            if not include_keys or prop_name in include_keys:
                _dict[prop_name] = prop_value

        return _dict



"""Exceptions module for ORM package"""
import traceback
from typing import Any, Dict, Iterable, Optional

# Reexport
from psycopg2.errors import InvalidTextRepresentation  # noqa: F401

# This is copied from ns_python_core for convenience but definitely belongs
# somewhere else.
ERROR_INFO_KEYS = [
    "internal_status_msg",
    "external_status_msg",
    "exception_type",
    "stack_trace",
]


class ErrorInfo(dict):
    """Error information that is passed between services.

    This class verifies that parameters are named correctly and can be
    serialized by the default JSON encoder to the correct format.
    """

    def _validate_keys(self, keys: Iterable[str]) -> None:
        """Validate that the given key is one of the expected keys.

        Will raise an exception if an unexpected key is encountered.
        """
        for key in keys:
            if key not in ERROR_INFO_KEYS:
                raise ValueError("Invalid key: '%s'" % key)

    def _ensure_external_status_message(self) -> None:
        """Make sure there is an external status message.

        Set a generic external status message if an internal status message
        is specified and no external status message is specified.
        """
        if (
            self["internal_status_msg"] is not None
            and self["external_status_msg"] is None
        ):
            self[
                "external_status_msg"
            ] = "An error occurred, please contact Narrative Science."

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Create a new instance of the ErrorInfo class

        This is initialized the same way a dict is.
        Keys that are not specified will receive a value of None.
        """
        for key in ERROR_INFO_KEYS:
            if key not in self:
                self[key] = None
        self.update(*args, **kwargs)

    def __setitem__(self, key: str, value: Any) -> None:
        """Update or set the value of an item.

        This will raise an exception if an unexpected key is passed in.
        """
        self._validate_keys([key])
        super(ErrorInfo, self).__setitem__(key, value)
        self._ensure_external_status_message()

    def update(self, *args: Any, **kwargs: Any) -> None:
        """Update the values in the dict.

        This will raise an exception if an unexpected value is passed in.
        """
        self._validate_keys(kwargs.keys())

        updated_args = []
        for arg in args:
            self._validate_keys(arg.keys())
            updated_args.append(arg)
        super(ErrorInfo, self).update(*updated_args, **kwargs)
        self._ensure_external_status_message()


class BaseException(Exception):
    """Base class for exceptions.

    Its only responsibility is to capture internal and external status
    information as well as a stack trace.
    """

    # This property denotes the component that owns will raise this exception. This
    # property can be used for component level monitoring
    COMPONENT = None

    def __init__(
        self,
        internal_msg: str = None,
        external_msg: str = None,
        error_info: ErrorInfo = None,
    ) -> None:
        """Initialize the exception.

        The error_info parameter is expected to be an ErrorInfo object or a
        dictionary with the same keys as an ErrorInfo object.
        If an error_info object is passed in the content of that object will
        override anything else in this class.
        """
        Exception.__init__(self, internal_msg)
        # Set internal message to '' by default to avoid TypeErrors
        self.internal_msg = internal_msg
        if internal_msg is None:
            self.internal_msg = ""
        self.error_info = error_info
        self.external_msg = external_msg

        # Attempt to capture a stack trace.  If an exception has occurred we'll
        # just grab that stack trace.  If one has not occurred we'll grab the
        # current stack minus the last frame which is just the constructor of
        # the exception class.
        self.stack_trace = None
        try:
            self.stack_trace = traceback.format_exc()
            if self.stack_trace == "None\n":
                self.stack_trace = "".join(traceback.format_stack()[:-1])
        except AttributeError:
            # Not in an exception context
            pass

    def get_error_info(self) -> ErrorInfo:
        """Build and return an ErrorInfo object

        The object should contain the error information collected by the
        exception class.  If an ErrorInfo object was passed in to the
        constructor, then return that instead.
        """
        if self.error_info is not None:
            return self.error_info
        else:
            return ErrorInfo(
                internal_status_msg=self.internal_msg,
                external_status_msg=self.external_msg,
                exception_type=self.__class__.__name__,
                stack_trace=self.stack_trace,
            )

    def get_metadata(self) -> Dict[str, str]:
        """Return the standard metadata for this error

        Returns:
            A dictionary of property keys to their stringified values

        """
        return {"component": self.COMPONENT.name}

    def __str__(self) -> str:
        """Returns the message describing the exception"""
        return self.internal_msg


class DatabaseRecordNotFound(BaseException):
    """Exception thrown when trying to fetch a database record that doesn't exist"""

    def __init__(self, table_name: str, **kwargs: Any) -> None:
        """Initialize DatabaseRecordNotFound

        Args:
            table_name: The name of the table that the record would be from
            kwargs: The field names and field values of the record that's not found

        """
        msg = (
            f"No record was found with the following fields: {kwargs} "
            f"on table '{table_name}'"
        )

        super().__init__(msg, msg)


class DatabaseObjectMisconfigured(BaseException):
    """Exception thrown when using a misconfigured DatabaseObject object"""

    def __init__(self, class_name: str) -> None:
        """Initialize DatabaseObjectMisconfigured

        Args:
            class_name: The name of the DatabaseObject class

        """
        msg = f"DatabaseObject class '{class_name}' is misconfigured"
        super().__init__(msg, msg)


class DatabaseModelMisconfigured(BaseException):
    """Exception thrown when using a misconfigured DatabaseModel object"""

    def __init__(self, class_name: str) -> None:
        """Initialize DatabaseModelMisconfigured

        Args:
            class_name: The name of the DatabaseModel class

        """
        msg = f"DatabaseModel '{class_name}' is misconfigured"
        super().__init__(msg, msg)


class DatabaseObjectMissingField(BaseException):
    """Exception thrown when the field provided for querying doesn't exist on the table"""

    def __init__(self, class_name: str, field_name: str) -> None:
        """Initialize DatabaseObjectMissingField

        Args:
            class_name: The name of the DatabaseObject class
            field_name: The name of the field used for querying this class

        """
        msg = f"DatabaseObject class '{class_name}' does not have field '{field_name}'"
        super().__init__(msg, msg)


class DatabaseModelMissingField(BaseException):
    """Exception thrown when the field provided for querying doesn't exist on the table"""

    def __init__(self, class_name: str, field_name: str) -> None:
        """Initialize DatabaseModelMissingField

        Args:
            class_name: The name of the DatabaseModel class
            field_name: The name of the field used for querying this class

        """
        msg = f"DatabaseModel class '{class_name}' does not have field '{field_name}'"
        super().__init__(msg, msg)


class InvalidMethodParameterization(BaseException):
    """Exception thrown when passing invalid parameters into a method"""

    def __init__(self, method_name: str, **kwargs: Any) -> None:
        """Initialize InvalidMethodParameterization

        Args:
            method_name: The name of the method
            kwargs: The parameters that were passed into the method

        """
        msg = (
            f"An invalid parameter configuration: {kwargs} "
            f"was passed in for the method '{method_name}'"
        )
        super().__init__(msg, msg)


class InvalidFieldValue(BaseException):
    """Exception thrown when a field value is invalid"""

    def __init__(
        self,
        field_name: Optional[str] = None,
        field_value: Optional[Any] = None,
        message: Optional[str] = None,
    ) -> None:
        """Initialize exception.

        Can create with the invalid field and field value or a prewritten message
        returned from postgres.

        Args:
            field_name: The name of the field with the invalid value
            field_value: The value that was invalid
            message: A pre-created message explaining the issue

        """
        if message:
            msg = message
        else:
            self.field_name = field_name
            self.field_value = field_value

            value_type = type(field_value)
            try:
                string_value = str(field_value)
            except Exception:
                string_value = "<couldn't convert>"

            msg = (
                f"An invalid value was used for field {field_name}: {string_value} "
                f"of type {value_type}."
            )

        super().__init__(msg, msg)


class InvalidSqlIdentifierErr(Exception):
    """Indicates a sql identifier is invalid"""

    def __init__(self, identifier: str) -> None:
        """Initialize InvalidSqlIndentifierErr

        Args:
            identifier: The invalid identifier

        """
        self.identifier = identifier

    def __str__(self) -> str:
        """Returns the message describing the exception"""
        return f"Invalid identifier {self.identifier}"


class NestedDatabaseModelNotResolved(BaseException):
    """Indicates a property was accessed before the reference was resolved"""

    def __init__(self, model_cls: str, nested_model_id_value: Any) -> None:
        """Initialize NestedDatabaseModelNotResolved

        Args:
            model_cls: The class name of the model that was being referenecd
            nested_model_id_value: The value of the unique id for this nested model

        """
        msg = (
            f"Object {model_cls} with id {nested_model_id_value} was not resolved."
            f" Please call `fetch()` before trying to access properties of {model_cls}"
        )

        super().__init__(msg, msg)



"""Functions for getting database engines and connections"""
import asyncio
from enum import Enum
from functools import wraps
import logging
import ssl
from typing import Any, Callable, Dict, NamedTuple, Optional, Tuple, Union
from urllib.parse import parse_qs

from aiopg import Connection as AIOPGConnection
from aiopg.sa import create_engine, Engine
from asyncpg.connection import Connection as AsyncPGConnection
from asyncpg.exceptions import ConnectionDoesNotExistError
from asyncpg.pool import Pool
from asyncpgsa import create_pool
import backoff
from sqlalchemy import create_engine as create_engine_sync
from sqlalchemy.engine import Engine as SyncEngine

from pynocular.aiopg_transaction import (
    ConditionalTransaction,
    transaction as Transaction,
)
from pynocular.config import DB_POOL_MAX_SIZE, DB_POOL_MIN_SIZE, POOL_RECYCLE

logger = logging.getLogger(__name__)

_engines: Dict[Tuple[str, str], Engine] = {}
_pools: Dict[Tuple[str, str], Pool] = {}


async def get_aiopg_engine(
    conn_str: str,
    enable_hstore: bool = True,
    force: bool = False,
    application_name: str = None,
    if_exists: bool = False,
) -> Optional[Engine]:
    """Returns the aiopg SQLAlchemy connection engine for a given connection string.

    This function lazily creates the connection engine if it doesn't already
    exist. Callers of this function shouldn't close the engine. It will be
    closed automatically when the process exits.

    This function exists to keep a single engine (and thus a single connection
    pool) per database, which prevents us from maxing out the number of
    connections the database server will give us.

    We include the hash of the event loop in the cache key because otherwise, if the
    event loop closes, the cached engine will raise an exception when it's used.

    Args:
        conn_str: The connection string for the engine
        enable_hstore: determines if the hstore should be enabled on the database.
            Redshift requires this to be disabled.
        force: Force the creation of the engine regardless of the cache
        application_name: Arbitrary string that shows up in queries to the
            ``pg_stat_activity`` view for tracking the source of database connections.
        if_exists: Only return the engine if it already exists

    Returns:
        aiopg engine for the connection string

    """
    global _engines
    logger.debug("Attempting to get DB engine")
    loop_hash = str(hash(asyncio.get_event_loop()))
    cache_key = (loop_hash, conn_str)
    engine = _engines.get(cache_key)

    if if_exists and engine is None:
        return None

    if engine is None or force or engine.closed:
        engine = await create_engine(
            conn_str,
            enable_hstore=enable_hstore,
            application_name=application_name,
            pool_recycle=POOL_RECYCLE,
        )
        _engines[cache_key] = engine
        logger.debug(f"DB engine created successfully: {engine}")

    logger.debug("DB engine retrieved")
    return engine


def get_psycopg_engine(
    conn_str: str, force: bool = False, if_exists: bool = False
) -> Optional[SyncEngine]:
    """Returns the psycopg SQLAlchemy connection engine for a given connection string.

    Similar to :py:func:`get_engine` except that a synchronous SQLAlchemy engine is
    created instead.

    Args:
        conn_str: The connection string for the engine
        force: Force the creation of the engine regardless of the cache
        if_exists: Only return the engine if it already exists

    Returns:
        SQLAlchemy engine for the connection string

    """
    global _engines
    logger.debug("Attempting to get sync DB engine")
    cache_key = ("sync", conn_str)
    engine = _engines.get(cache_key)

    if if_exists and engine is None:
        return None

    if engine is None or force:
        engine = create_engine_sync(conn_str, pool_recycle=POOL_RECYCLE)
        _engines[cache_key] = engine
        logger.debug(f"Sync DB engine created successfully: {engine}")

    logger.debug("Sync DB engine retrieved")
    return engine


@backoff.on_exception(
    backoff.expo, ConnectionDoesNotExistError, max_tries=8, jitter=backoff.random_jitter
)
async def get_asyncpgsa_pool(
    conn_str: str,
    force: bool = False,
    application_name: str = None,
    if_exists: bool = False,
) -> Optional[Pool]:
    """Returns the asyncpgsa connection pool for a given connection string.

    This function lazily creates the connection pool if it doesn't already
    exist. Callers of this function shouldn't close the pool. It will be
    closed automatically when the process exits.

    This function exists to keep a single pool per database, which prevents us
    from maxing out the number of connections the database server will give us.

    We include the hash of the event loop in the cache key because otherwise, if the
    event loop closes, the cached pool will raise an exception when it's used.

    Args:
        conn_str: The connection string for the engine
        force: Force the creation of the pool regardless of the cache
        application_name: Arbitrary string that shows up in queries to the
            ``pg_stat_activity`` view for tracking the source of database connections.
        if_exists: Only return the pool if it already exists

    Returns:
        The aiopgsa pool for that connection

    """
    global _pools
    logger.debug("Attempting to get/create the db pool")
    loop_hash = str(hash(asyncio.get_event_loop()))
    cache_key = (loop_hash, conn_str)
    pool = _pools.get(cache_key)

    # Split the query params from the connection string
    parsed_conn = conn_str.split("?")
    conn_str = parsed_conn[0]
    ssl_cert_path = None

    if len(parsed_conn) > 1:
        query_params = parse_qs(parsed_conn[1])
        if "sslrootcert" in query_params:
            ssl_cert_path = query_params["sslrootcert"][0]

    if if_exists and pool is None:
        return None

    if pool is None or force or pool._closed:
        options = {}
        if application_name is not None:
            options["server_settings"] = {"application_name": application_name}
        if ssl_cert_path is not None:
            ssl_context = ssl.create_default_context()
            ssl_context.load_verify_locations(cafile=ssl_cert_path)
            options["ssl"] = ssl_context  # type: ignore
        try:
            pool = await create_pool(
                conn_str,
                min_size=DB_POOL_MIN_SIZE,
                max_size=DB_POOL_MAX_SIZE,
                max_inactive_connection_lifetime=POOL_RECYCLE,
                **options,
            )
        except Exception as e:
            logger.error(
                {
                    "description": "Failed to create asyncpg connection pool",
                    "application_name": application_name,
                    "min_size": DB_POOL_MIN_SIZE,
                    "max_size": DB_POOL_MAX_SIZE,
                    "force": force,
                    "if_exists": if_exists,
                    "error_msg": repr(e),
                }
            )
            raise e

        _pools[cache_key] = pool
        logger.debug(f"Created new asyncpg pool: {pool}")

    logger.debug("Connection pool successfully retrieved")
    return pool


class DatabaseType(Enum):
    """Database type to differentiate engines and pools"""

    aiopg_engine = "aiopg_engine"
    asyncpgsa_pool = "asyncpgsa_pool"
    sqlalchemy_engine = "sqlalchemy_engine"


class DBInfo(NamedTuple):
    """Data class for a database's connection information"""

    engine_type: DatabaseType
    connection_string: str
    enable_hstore: bool = True


class DBEngine:
    """Wrapper over database engine types"""

    @classmethod
    async def _get_engine_or_pool(
        cls,
        db_info: DBInfo,
        force: bool = False,
        application_name: str = None,
        if_exists: bool = False,
    ) -> Optional[Union[Engine, SyncEngine, Pool]]:
        """Get an aiopg engine or asyncpg pool depending on the database configuration.

        Args:
            db_info: Information for making the database connection
            force: Force the creation of the pool regardless of the cache
            application_name: Arbitrary string that shows up in queries to the
                ``pg_stat_activity`` view for tracking the source of database
                connections.
            if_exists: Only return the engine or pool if it already exists

        Returns:
            database engine or pool

        Raises:
            :py:exec:`ValueError` if the database type isn't supported

        """
        if db_info.engine_type == DatabaseType.aiopg_engine:
            return await get_aiopg_engine(
                db_info.connection_string,
                enable_hstore=db_info.enable_hstore,
                force=force,
                application_name=application_name,
                if_exists=if_exists,
            )
        elif db_info.engine_type == DatabaseType.asyncpgsa_pool:
            return await get_asyncpgsa_pool(
                db_info.connection_string,
                force=force,
                application_name=application_name,
                if_exists=if_exists,
            )
        elif db_info.engine_type == DatabaseType.sqlalchemy_engine:
            return get_psycopg_engine(
                db_info.connection_string, force=force, if_exists=if_exists
            )

        raise ValueError(f"Unsupported database type: {db_info.engine_type}")

    @classmethod
    async def get_engine(
        cls, db_info: DBInfo, force: bool = False, application_name: str = None
    ) -> Union[Engine, SyncEngine]:
        """Get a SQLAlchemy connection engine for a given database alias.

        See :py:func:`.get_engine` for more details.

        Args:
            db_info: database connection information
            force: Force the creation of the pool regardless of the cache
            application_name: Arbitrary string that shows up in queries to the
                ``pg_stat_activity`` view for tracking the source of database
                connections.

        Returns:
            database engine

        """
        return await cls._get_engine_or_pool(
            db_info, force=force, application_name=application_name
        )

    @classmethod
    async def get_pool(
        cls, db_info: DBInfo, force: bool = False, application_name: str = None
    ) -> Pool:
        """Get a asyncpgsa connection pool for a given database alias.

        See :py:func:`.get_pool` for more details.

        Args:
            db_info: database connection information
            force: Force the creation of the pool regardless of the cache
            application_name: Arbitrary string that shows up in queries to the
                ``pg_stat_activity`` view for tracking the source of database
                connections.

        Returns:
            database connection pool

        """
        if db_info.engine_type != DatabaseType.asyncpgsa_pool:
            raise ValueError(f"Cannot get a connection pool with {db_info.engine_type}")
        return await cls._get_engine_or_pool(
            db_info, force=force, application_name=application_name
        )

    @classmethod
    async def acquire(
        cls, db_info: DBInfo
    ) -> Union[AIOPGConnection, AsyncPGConnection]:
        """Acquire a SQLAlchemy connection for a given database alias.

        This is a convenience function that first gets/creates the engine or pool then
        calls acquire. This returns a context manager.

        Args:
            db_info: database connection information

        Returns:
            context manager that yields the connection

        """
        engine = await cls._get_engine_or_pool(db_info)
        return engine.acquire()

    @classmethod
    async def transaction(
        cls, db_info: DBInfo, is_conditional: bool = False
    ) -> Union[ConditionalTransaction, Transaction]:
        """Acquire a SQLAlchemy transaction for a given database alias.

        This is a convenience function that first gets/creates the engine then calls
        ConditionalTransaction. This returns a context manager.

        Args:
            db_info: database connection information
            is_conditional: If true, returns a conditional transaction.

        Returns:
            Transaction or ConditionalTransaction for use as a context manager

        """
        if db_info.engine_type != DatabaseType.aiopg_engine:
            raise ValueError(
                f"Transaction does not support database type {db_info.engine_type}"
            )
        engine = await cls._get_engine_or_pool(db_info)
        return ConditionalTransaction(engine) if is_conditional else Transaction(engine)

    @classmethod
    def open_transaction(cls, db_info: DBInfo) -> Callable:
        """Decorator that wraps the function call in a database transaction

        Args:
            database_alias: The database alias to use for the transaction

        Returns:
            The wrapped function call

        """

        def parameterized_decorator(fn: Callable) -> Callable:
            """Function that will create the wrapper function

            Args:
                fn: The function to wrap

            Returns:
                The wrapped function

            """

            @wraps(fn)
            async def wrapped_funct(*args: Any, **kwargs: Any) -> Any:
                """The actual wrapper function

                Args:
                    args: The argument calls to the wrapped function
                    kwargs: The keyword args to the wrapped function

                Returns:
                    The result of the function

                """
                async with await DBEngine.transaction(db_info, is_conditional=False):
                    ret = await fn(*args, **kwargs)
                return ret

            return wrapped_funct

        return parameterized_decorator

    @classmethod
    async def close(cls, db_info: DBInfo) -> None:
        """Close existing database engines and pools

        Args:
            db_info: database connection information

        """
        logger.info("Closing database engine")
        pool_engine = await cls._get_engine_or_pool(db_info, if_exists=True)
        if pool_engine is None:
            # The engine/pool doesn't exist so nothing to close
            pass
        elif db_info.engine_type == DatabaseType.asyncpgsa_pool:
            await pool_engine.close()
        elif db_info.engine_type == DatabaseType.sqlalchemy_engine:
            pool_engine.dispose()
        else:
            pool_engine.close()
            await pool_engine.wait_closed()



"""Module for aiopg transaction utils"""
import asyncio
import sys
from typing import Dict, Optional, Union

import aiocontextvars as contextvars
from aiopg.sa.connection import SAConnection
import aiopg.sa.engine

transaction_connections_var = contextvars.ContextVar(
    "transaction_connections", default={}
)


def get_current_task() -> asyncio.Task:
    """Get the current task when this method is called

    Returns:
        The current task the method is called in

    """
    if sys.version_info.major == 3 and sys.version_info.minor > 7:
        # If this is version 3.7 or higher then use the new function to get the current task
        return asyncio.current_task()
    else:
        return asyncio.Task.current_task()


class LockedConnection(SAConnection):
    """A wrapper connection class that won't make multiple queries at once"""

    def __init__(self, connection: SAConnection) -> None:
        """Create a new LockedConnection

        Args:
            connection: The connection to wrap

        """
        self._conn = connection
        self.lock = asyncio.Lock()

    async def execute(self, *args, **kwargs):
        """Wrapper around the `execute` method of the wrapped SAConnection"""
        async with self.lock:
            return await self._conn.execute(*args, **kwargs)

    def __getattr__(self, attr):
        """Except for execute, all other attributes should pass through"""
        return getattr(self._conn, attr)


class TaskContextConnection:
    """Interface for managing a connection entry on the asyncio Task context

    The current asyncio.Task has a context attribute that keeps track of various keys.
    We'll use this to store the open connection so we can perform our nested/conditional
    transaction logic in :py:class:`transaction`. The actual value stored on the context
    is a dict of connections keyed by the engine.
    """

    def __init__(self, connection_key: str) -> None:
        """Initializer

        Args:
            connection_key: Key for getting/setting/clearing from the connection map

        """
        self.connection_key = connection_key
        self._token: Optional[contextvars.Token] = None

        # Set the asyncio task context if it's not set already. We'll look in the
        # context for an open connection.
        task = get_current_task()
        if not hasattr(task, "context"):
            task.context = contextvars.copy_context()

    @classmethod
    def _get_connections(cls) -> Dict[str, LockedConnection]:
        """Get the map of connections from the task context"""
        global transaction_connections_var
        return transaction_connections_var.get()

    def get(self) -> Optional[LockedConnection]:
        """If there is already a connection stored, get it"""
        return self._get_connections().get(self.connection_key)

    def set(self, conn: LockedConnection) -> contextvars.Token:
        """Set the connection on the context

        Args:
            conn: Connection to store

        Returns:
            contextvars token used to reset the var in :py:meth:`.clear`

        """
        global transaction_connections_var
        connections = self._get_connections()
        connections[self.connection_key] = conn
        token = transaction_connections_var.set(connections)
        self._token = token
        return token

    def clear(self) -> None:
        """Clear the connection from the context"""
        if not self._token:
            raise ValueError("Token must be defined")

        global transaction_connections_var
        transaction_connections_var.reset(self._token)


class transaction:
    """A context manager to collect nested calls in a transaction

    To use, anywhere you want to have queries put into a transaction, do

        async with transaction(aiopg_engine) as trx:
            ...

    The resulting trx object can be used just like a connection you would
    get from `aiopg_engine.acquire()`, but any nested usages of this decorator
    will ensure that we do not deadlock from nested acquire calls, and do not
    run into errors where we attempt to use the same connection to make
    multiple calls at once.
    NB: It does this by ensuring that we get the same connection object and
    execute serially, so you only want to use this in cases where you are
    worried about these issues.

    For example, using just aiopg, this is an error:

        async with engine.acquire() as conn:
            await asyncio.gather(
                conn.execute(TABLE.insert().values(id=uuid(), name="foo")),
                conn.execute(TABLE.insert().values(id=uuid(), name="bar")))

    But using this class, we will not error:

        async with transaction(engine) as conn:
            await asyncio.gather(
                conn.execute(TABLE.insert().values(id=uuid(), name="foo")),
                conn.execute(TABLE.insert().values(id=uuid(), name="bar")))

    Note:
        There are limits to the transaction rollback protection that this
        context manager affords. Specifically, a known failure case can be
        encountered if a DB connection is created by calling `Engine.acquire`
        rather than `transaction(Engine)`, even if the call to `acquire` is
        made within a transaction context. For more information, see:
        :py:module:`python_core.tests.functional
        .test_aiopg_transaction_integrity`.

    """

    def __init__(self, engine: aiopg.sa.engine.Engine) -> None:
        """Create a new transaction context

        Args:
            engine: Database engine for making connections

        """
        self._engine = engine

        # Is this the outer-most transaction context?
        # If so, this will be set to true in `__aenter__`
        self._top = False

        # If we have started a transaction, store it here
        self._trx = None

        # Initiatize an interface for managing the connection on the asyncio task context
        self.task_connection = TaskContextConnection(str(engine))

    async def __aenter__(self) -> LockedConnection:
        """Establish the transaction context

        Figure out if this is the top level context. If so, get a connection
        and start a transaction. If not, then just grab the stored connection.
        """
        conn = self.task_connection.get()
        if not conn:
            # There is no stored connection in this context, so this must be
            # the top level call.
            self._top = True
            # Create the connection
            conn = LockedConnection(await self._engine.acquire())
            self.task_connection.set(conn)
            # Start a transaction
            try:
                self._trx = await conn.begin()
            except Exception:
                await conn.close()
                self.task_connection.clear()
                raise
        return conn

    async def __aexit__(self, exc_type, exc_value, tb) -> None:
        """Exit the transaction context

        If this is the top level context, then commit the transaction (unless
        there was an error, in which case we should rollback instead).
        If this is not the top level context, we don't need to do anything,
        since everything will be committed or rolled back by that top level
        context.
        """
        if self._top:
            # We may have gotten here from an error, in which case it is
            # possible that we are also awaiting for a query to finish
            # executing. So before rolling back the connection, make sure we
            # can acquire the connection lock to ensure nothing else is
            # executing
            conn = self.task_connection.get()
            async with conn.lock:
                try:
                    if exc_type:  # There was an exception
                        await self._trx.rollback()
                    else:
                        await self._trx.commit()
                finally:
                    self.task_connection.clear()
                    await conn.close()


class ConditionalTransaction(transaction):
    """Context manager to conditionally collect nested calls in a transaction

    This context manager allows you to conditionally execute code in a
    transaction if nested within another transaction. If it is the top level
    "transaction", this will behave like a standard `engine.acquire()`. Usage
    is otherwise the same as for the parent transaction class.

    Examples:
        This Will behave the same as `engine.acquire`, assuming this is not
        nested under a transaction elsewhere

            async with ConditionalTransaction(engine) as trx:
                ...

        This will behave as a nested transaction:

            async with transaction(engine) as outer_trx:
                async with ConditionalTransaction(engine) as inner_trx:
                    ...

    """

    def __init__(self, engine: aiopg.sa.engine.Engine) -> None:
        """Initialize the context manager

        Args:
            engine: An aiopg engine

        """
        super().__init__(engine)
        # The connection object, if functioning as standard connection
        self._conn = None

    async def __aenter__(self) -> Union[LockedConnection, SAConnection]:
        """Conditionally establish the transaction context

        Returns:
            Either a locked connection or a standard connection, depending on
            whether this context manager is nested under a transaction.

        """
        conn = self.task_connection.get()
        # If there is already a connection stored, act as a transaction
        if conn:
            return await super().__aenter__()
        # Otherwise behave as a standard connection
        self._conn = await self._engine.acquire()
        return self._conn

    async def __aexit__(self, exc_type, exc_value, tb) -> None:
        """Exit the transaction context"""
        if self._conn is not None:
            await self._conn.close()
        else:
            await super().__aexit__(exc_type, exc_value, tb)



"""Database utility functions"""

import logging
import re

from aiopg.sa.connection import SAConnection
import sqlalchemy as sa
from sqlalchemy.sql.ddl import CreateTable

from pynocular.engines import DatabaseType, DBEngine, DBInfo
from pynocular.exceptions import InvalidSqlIdentifierErr

logger = logging.getLogger()


async def create_new_database(connection_string: str, db_name: str) -> None:
    """Create a new database database for testing

    Args:
        connection_string: A connection string for the database
        db_name: the name of the database to create

    """
    existing_db = DBInfo(DatabaseType.aiopg_engine, connection_string)
    conn = await (await DBEngine.get_engine(existing_db)).acquire()
    # End existing commit
    await conn.execute("commit")
    # Create db
    await conn.execute(f"drop database if exists {db_name}")
    await conn.execute(f"create database {db_name}")
    await conn.close()


async def create_table(db_info: DBInfo, table: sa.Table) -> None:
    """Create table in database

    Args:
        db_info: Information for the database to connect to
        table: The table to create

    """
    engine = await DBEngine.get_engine(db_info)
    conn = await engine.acquire()
    await conn.execute(CreateTable(table))
    await conn.close()


async def drop_table(db_info: DBInfo, table: sa.Table) -> None:
    """Drop table in database

    Args:
        db_info: Information for the database to connect to
        table: The table to create

    """
    engine = await DBEngine.get_engine(db_info)
    conn = await engine.acquire()
    await conn.execute(f"drop table if exists {table.name}")
    await conn.close()


async def setup_datetime_trigger(conn: SAConnection) -> None:
    """Set up created_at/updated_at datetime trigger

    Args:
        conn: an async sqlalchemy connection

    """
    await conn.execute('CREATE EXTENSION IF NOT EXISTS "uuid-ossp";')
    await conn.execute('CREATE EXTENSION IF NOT EXISTS "plpgsql";')
    await conn.execute(
        """
        CREATE OR REPLACE FUNCTION update_timestamp_columns()
        RETURNS TRIGGER AS $$
        BEGIN
            IF NEW.created_at IS NULL THEN
                NEW.created_at = now();
            END IF;

            NEW.updated_at = now();
            RETURN NEW;
        END;
        $$ language 'plpgsql';
        """
    )


async def add_datetime_trigger(conn: SAConnection, table: str) -> None:
    """Helper method for adding created_at and updated_at datetime triggers on a table


    Args:
        conn: an async sqlalchemy connection
        table: The name of the table to add an edit trigger for

    """
    await setup_datetime_trigger(conn)
    await conn.execute(
        """
        CREATE TRIGGER update_{table}_timestamps
        BEFORE INSERT OR UPDATE ON {table}
        FOR EACH ROW EXECUTE PROCEDURE update_timestamp_columns();
    """.format(
            table=table
        )
    )


async def remove_datetime_trigger(conn: SAConnection, table: str) -> None:
    """Helper method for removing datetime triggers on a table

    Args:
        conn: an async sqlalchemy connection
        table: The name of the table to remove a trigger for

    """
    await conn.execute(
        "DROP TRIGGER IF EXISTS update_{table}_timestamps on {table}".format(
            table=table
        )
    )


def get_cleaned_db_name(
    name: str,
    lowercase: bool = True,
    remove_leading_numbers: bool = True,
    replace_spaces_with_underscores: bool = True,
    replace_dashes_with_underscores: bool = True,
    remove_special_chars: bool = True,
    limit: int = 128,
) -> str:
    """Gets a name cleaned to adhere to sql naming conventions

    Args:
        name: An uncleaned name (such as a table or column name)
        lowercase: Whether all letters in the name should be lowercased
        remove_leading_numbers: Whether leading numbers should be stripped
        replace_spaces_with_underscores: Whether spaces should be replaced with underscores
        replace_dashes_with_underscores: Whether dashes should be replaced with underscores
        remove_special_chars: Whether any characters other than letters, numbers, and
            underscores should be removed from the name
        limit: the maximum allowed length of the name after cleaning. The default value
            is the Athena/Glue column name length limit.

    Returns:
        A cleaned name to be used in a relational database

    Raises:
        :py:exc:`InvalidSqlIdentifierErr`: If the name is still invalid
            after being cleaned

    """
    cleaned_name = name

    if lowercase:
        cleaned_name = cleaned_name.lower()

    if remove_leading_numbers:
        cleaned_name = cleaned_name.lstrip("0123456789")

    if replace_spaces_with_underscores:
        cleaned_name = "_".join(cleaned_name.split(" "))

    if replace_dashes_with_underscores:
        cleaned_name = "_".join(cleaned_name.split("-"))

    if remove_special_chars:
        cleaned_name = re.sub(r"[^a-zA-Z0-9_]*", "", cleaned_name)

    if len(cleaned_name) == 0 or len(cleaned_name) > limit:
        raise InvalidSqlIdentifierErr(cleaned_name)

    return cleaned_name



"""Configuration for engines and models"""
import os

POOL_RECYCLE = int(os.environ.get("POOL_RECYCLE", 300))
DB_POOL_MIN_SIZE = int(os.environ.get("DB_POOL_MIN_SIZE", 2))
DB_POOL_MAX_SIZE = int(os.environ.get("DB_POOL_MAX_SIZE", 10))



"""Class that wraps nested DatabaseModels"""
from typing import Any, Callable

from pynocular.exceptions import NestedDatabaseModelNotResolved


class NestedDatabaseModel:
    """Class that wraps nested DatabaseModels"""

    def __init__(
        self,
        model_cls: Callable,
        _id: Any,
        model: "DatabaseModel" = None,  # noqa
    ) -> None:
        """Init for NestedDatabaseModel

        Args:
            model_cls: The class that the id relates to
            _id: The id of the references
            model: The model object if it is already loaded

        """
        self._model_cls = model_cls
        self._model = model
        # We can only support nested database models that are based off of a single
        # unique identifier
        self._primary_key_name = model_cls._primary_keys[0].name
        setattr(self, self._primary_key_name, _id)

    def get_primary_id(self) -> Any:
        """Standard interface for returning the id of a field

        Returns:
            The ID value for the proxied DatabaseModel

        """
        return getattr(self, self._primary_key_name)

    async def fetch(self) -> None:
        """Resolves the reference via the id set"""
        if self._model is None:
            self._model = await self._model_cls.get(
                getattr(self, self._primary_key_name)
            )

    def __getattr__(self, attr_name: str) -> Any:
        """Wrapper around getattr

        This will only get hit if the class doesn't have a reference to attr_name

        Args:
            attr_name: The name of the attribute

        Returns:
            The value of the attribute on the object

        """
        if self._model is None:
            raise NestedDatabaseModelNotResolved(self._model_cls, self.get_primary_id())
        else:
            return getattr(self._model, attr_name)

    def __eq__(self, other: Any) -> bool:
        """Equality function

        Args:
            other: The object to compare to

        Returns:
            If the object is equal to this one

        """

        if self._model is None:
            return False
        else:
            return self._model == other



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pynocular']

package_data = \
{'': ['*']}

install_requires = \
['aenum>=3.1.0,<4.0.0',
 'aiocontextvars>=0.2.2,<0.3.0',
 'aiopg[sa]>=1.3.1,<2.0.0',
 'asyncpg>=0.24.0,<0.25.0',
 'asyncpgsa>=0.24.0,<0.25.0',
 'backoff>=1.11.1,<2.0.0',
 'pydantic>=1.6,<2.0']

setup_kwargs = {
    'name': 'pynocular',
    'version': '0.11.0',
    'description': 'Lightweight ORM that lets you query your database using Pydantic models and asyncio',
    'long_description': '# pynocular\n\n[![](https://img.shields.io/pypi/v/pynocular.svg)](https://pypi.org/pypi/pynocular/) [![License](https://img.shields.io/badge/License-BSD%203--Clause-blue.svg)](https://opensource.org/licenses/BSD-3-Clause)\n\nPynocular is a lightweight ORM that lets you query your database using Pydantic models and asyncio.\n\nWith Pynocular you can decorate your existing Pydantic models to sync them with the corresponding table in your\ndatabase, allowing you to persist changes without ever having to think about the database. Transaction management is\nautomatically handled for you so you can focus on the important parts of your code. This integrates seamlessly with frameworks that use Pydantic models such as FastAPI.\n\nFeatures:\n\n- Fully supports asyncio to write to SQL databases\n- Provides simple methods for basic SQLAlchemy support (create, delete, update, read)\n- Contains access to more advanced functionality such as custom SQLAlchemy selects\n- Contains helper functions for creating new database tables\n- Advanced transaction management system allows you to conditionally put requests in transactions\n\nTable of Contents:\n\n- [Installation](#installation)\n- [Guide](#guide)\n  - [Basic Usage](#basic-usage)\n  - [Advanced Usage](#advanced-usage)\n  - [Creating database tables](#creating-database-tables)\n- [Development](#development)\n\n## Installation\n\npynocular requires Python 3.6 or above.\n\n```bash\npip install pynocular\n# or\npoetry add pynocular\n```\n\n## Guide\n\n### Basic Usage\n\nPynocular works by decorating your base Pydantic model with the function `database_model`. Once decorated\nwith the proper information, you can proceed to use that model to interface with your specified database table.\n\nThe first step is to define a `DBInfo` object. This will contain the connection information to your database.\n\n```python\nfrom pynocular.engines import DatabaseType, DBInfo\n\n\n# Example below shows how to connect to a locally-running Postgres database\nconnection_string = f"postgresql://{db_user_name}:{db_user_password}@localhost:5432/{db_name}?sslmode=disable"\n)\ndb_info = DBInfo(DatabaseType.aiopg_engine, connection_string)\n```\n\nPynocular supports connecting to your database through two different asyncio engines; aiopg and asyncpgsa.\nYou can pick which one you want to use by passing the correct `DatabaseType` enum value into `DBInfo`.\n\n#### Object Management\n\nOnce you define a `db_info` object, you are ready to decorate your Pydantic models and interact with your database!\n\n```python\nfrom pydantic import BaseModel, Field\nfrom pynocular.database_model import database_model, UUID_STR\n\nfrom my_package import db_info\n\n@database_model("organizations", db_info)\nclass Org(BaseModel):\n\n    id: Optional[UUID_STR] = Field(primary_key=True, fetch_on_create=True)\n    name: str = Field(max_length=45)\n    slug: str = Field(max_length=45)\n    tag: Optional[str] = Field(max_length=100)\n\n    created_at: Optional[datetime] = Field(fetch_on_create=True)\n    updated_at: Optional[datetime] = Field(fetch_on_update=True)\n\n#### Object management\n\n# Create a new Org via `create`\norg = await Org.create(name="new org", slug="new-org")\n\n\n# Create a new Org via `save`\norg2 = Org(name="new org2", slug="new-org2")\nawait org2.save()\n\n\n# Update an org\norg.name = "renamed org"\nawait org.save()\n\n\n# Delete org\nawait org.delete()\n\n\n# Get org\norg3 = await Org.get(org2.id)\nassert org3 == org2\n\n# Get a list of orgs\norgs = await Org.get_list()\n\n# Get a filtered list of orgs\norgs = await Org.get_list(tag="green")\n\n# Get orgs that have several different tags\norgs = await Org.get_list(tag=["green", "blue", "red"])\n\n# Fetch the latest state of a table in the db\norg3.name = "fake name"\nawait org3.fetch()\nassert org3.name == "new org2"\n\n```\n\n#### Serialization\n\nDatabaseModels have their own serialization functions to convert to and from\ndictionaries.\n\n```python\n# Serializing org with `to_dict()`\norg = Org.create(name="org serialize", slug="org-serialize")\norg_dict = org.to_dict()\nexpected_org_dict = {\n    "id": "e64f6c7a-1bd1-4169-b482-189bd3598079",\n    "name": "org serialize",\n    "slug": "org-serialize",\n    "created_at": "2018-01-01 7:03:45",\n    "updated_at": "2018-01-01 9:24:12"\n}\nassert org_dict == expected_org_dict\n\n\n# De-serializing org with `from_dict()`\nnew_org = Org.from_dict(expected_org_dict)\nassert org == new_org\n```\n\n#### Using Nested DatabaseModels\n\nPynocular also supports basic object relationships. If your database tables have a\nforeign key reference you can leverage that in your pydantic models to increase the\naccessibility of those related objects.\n\n```python\nfrom pydantic import BaseModel, Field\nfrom pynocular.database_model import database_model, nested_model, UUID_STR\n\nfrom my_package import db_info\n\n@database_model("users", db_info)\nclass User(BaseModel):\n\n    id: Optional[UUID_STR] = Field(primary_key=True, fetch_on_create=True)\n    username: str = Field(max_length=100)\n\n    created_at: Optional[datetime] = Field(fetch_on_create=True)\n    updated_at: Optional[datetime] = Field(fetch_on_update=True)\n\n@database_model("organizations", db_info)\nclass Org(BaseModel):\n\n    id: Optional[UUID_STR] = Field(primary_key=True, fetch_on_create=True)\n    name: str = Field(max_length=45)\n    slug: str = Field(max_length=45)\n    # `organizations`.`tech_owner_id` is a foreign key to `users`.`id`\n    tech_owner: Optional[nested_model(User, reference_field="tech_owner_id")]\n    # `organizations`.`business_owner_id` is a foreign key to `users`.`id`\n    business_owner: nested_model(User, reference_field="business_owner_id")\n    tag: Optional[str] = Field(max_length=100)\n\n    created_at: Optional[datetime] = Field(fetch_on_create=True)\n    updated_at: Optional[datetime] = Field(fetch_on_update=True)\n\n\ntech_owner = await User.create("tech owner")\nbusiness_owner = await User.create("business owner")\n\n\n# Creating org with only business owner set\norg = await Org.create(\n    name="org name",\n    slug="org-slug",\n    business_owner=business_owner\n)\n\nassert org.business_owner == business_owner\n\n# Add tech owner\norg.tech_owner = tech_owner\nawait org.save()\n\n# Fetch from the db and check ids\norg2 = Org.get(org.id)\nassert org2.tech_owner.id == tech_owner.id\nassert org2.business_owner.id == business_owner.id\n\n# Swap user roles\norg2.tech_owner = business_owner\norg2.business_owner = tech_owner\nawait org2.save()\norg3 = await Org.get(org2.id)\nassert org3.tech_owner.id == business_owner.id\nassert org3.business_owner.id == tech_owner.id\n\n\n# Serialize org\norg_dict = org3.to_dict()\nexpected_org_dict = {\n    "id": org3.id,\n    "name": "org name",\n    "slug": "org-slug",\n    "business_owner_id": tech_owner.id,\n    "tech_owner_id": business_owner.id,\n    "tag": None,\n    "created_at": org3.created_at,\n    "updated_at": org3.updated_at\n}\n\nassert org_dict == expected_org_dict\n\n```\n\nWhen using `DatabaseModel.get(..)`, any foreign references will need to be resolved before any properties besides the primary ID can be accessed. If you try to access a property before calling `fetch()` on the nested model, a `NestedDatabaseModelNotResolved` error will be thrown.\n\n```python\norg_get = await Org.get(org3.id)\norg_get.tech_owner.id # Does not raise `NestedDatabaseModelNotResolved`\norg_get.tech_owner.username # Raises `NestedDatabaseModelNotResolved`\n\norg_get = await Org.get(org3.id)\nawait org_get.tech_owner.fetch()\norg_get.tech_owner.username # Does not raise `NestedDatabaseModelNotResolved`\n```\n\nAlternatively, calling `DatabaseModel.get_with_refs()` instead of `DatabaseModel.get()` will\nautomatically fetch the referenced records and fully resolve those objects for you.\n\n```python\norg_get_with_refs = await Org.get_with_refs(org3.id)\norg_get_with_refs.tech_owner.username # Does not raise `NestedDatabaseModelNotResolved`\n```\n\nThere are some situations where none of the objects have been persisted to the\ndatabase yet. In this situation, you can call `Database.save(include_nested_models=True)`\non the object with the references and it will persist all of them in a transaction.\n\n```python\n# We create the objects but dont persist them\ntech_owner = User("tech owner")\nbusiness_owner = User("business owner")\n\norg = Org(\n    name="org name",\n    slug="org-slug",\n    business_owner=business_owner\n)\n\nawait org.save(include_nested_models=True)\n```\n\n#### Special Type arguments\n\nWith Pynocular you can set fields to be optional and set by the database. This is useful\nif you want to let the database autogenerate your primary key or `created_at` and `updated_at` fields\non your table. To do this you must:\n\n- Wrap the typehint in `Optional`\n- Provide keyword arguments of `fetch_on_create=True` or `fetch_on_update=True` to the `Field` class\n\n### Advanced Usage\n\nFor most use cases, the basic usage defined above should suffice. However, there are certain situations\nwhere you don\'t necessarily want to fetch each object or you need to do more complex queries that\nare not exposed by the `DatabaseModel` interface. Below are some examples of how those situations can\nbe addressed using Pynocular.\n\n#### Tables with compound keys\n\nPynocular supports tables that use multiple fields as its primary key such as join tables.\n\n```python\nfrom pydantic import BaseModel, Field\nfrom pynocular.database_model import database_model, nested_model, UUID_STR\n\nfrom my_package import db_info\n\n@database_model("user_subscriptions", db_info)\nclass UserSubscriptions(BaseModel):\n\n    user_id: UUID_STR = Field(primary_key=True, fetch_on_create=True)\n    subscription_id: UUID_STR = Field(primary_key=True, fetch_on_create=True)\n    name: str\n\n\nuser_sub = await UserSub.create(\n    user_id="4d4254c4-8e99-45f9-8261-82f87991c659",\n    subscription_id="3cc5d476-dbe6-4cc1-9390-49ebd7593a3d",\n    name="User 1\'s subscriptions"\n)\n\n# Get the users subscription and confirm its the same\nuser_sub_get = await UserSub.get(\n    user_id="4d4254c4-8e99-45f9-8261-82f87991c659",\n    subscription_id="3cc5d476-dbe6-4cc1-9390-49ebd7593a3d",\n)\nassert user_sub_get == user_sub\n\n# Change a property value like any other object\nuser_sub_get.name = "change name"\nawait user_sub_get.save()\n```\n\n#### Batch operations on tables\n\nSometimes you want to insert a bunch of records into a database and you don\'t want to do an insert for each one.\nThis can be handled by the `create_list` function.\n\n```python\norg_list = [\n    Org(name="org1", slug="org-slug1"),\n    Org(name="org2", slug="org-slug2"),\n    Org(name="org3", slug="org-slug3"),\n]\nawait Org.create_list(org_list)\n```\n\nThis function will insert all records into your database table in one batch.\n\nIf you have a use case that requires deleting a bunch of records based on some field value, you can use `delete_records`:\n\n```python\n# Delete all records with the tag "green"\nawait Org.delete_records(tag="green")\n\n# Delete all records with if their tag has any of the following: "green", "blue", "red"\nawait Org.delete_records(tag=["green", "blue", "red"])\n```\n\nSometimes you may want to update the value of a record in a database without having to fetch it first. This can be accomplished by using\nthe `update_record` function:\n\n```python\nawait Org.update_record(\n    id="05c0060c-ceb8-40f0-8faa-dfb91266a6cf",\n    tag="blue"\n)\norg = await Org.get("05c0060c-ceb8-40f0-8faa-dfb91266a6cf")\nassert org.tag == "blue"\n```\n\n#### Complex queries\n\nSometimes your application will require performing complex queries, such as getting the count of each unique field value for all records in the table.\nBecause Pynocular is backed by SQLAlchemy, we can access table columns directly to write pure SQLAlchemy queries as well!\n\n```python\nfrom sqlalchemy import func, select\nfrom pynocular.engines import DBEngine\nasync def generate_org_stats():\n    query = (\n        select([func.count(Org.column.id), Org.column.tag])\n        .group_by(Org.column.tag)\n        .order_by(func.count().desc())\n    )\n    async with await DBEngine.transaction(Org._database_info, is_conditional=False) as conn:\n        result = await conn.execute(query)\n        return [dict(row) async for row in result]\n```\n\nNOTE: `DBengine.transaction` is used to create a connection to the database using the credentials passed in.\nIf `is_conditional` is `False`, then it will add the query to any transaction that is opened in the call chain. This allows us to make database calls\nin different functions but still have them all be under the same database transaction. If there is no transaction opened in the call chain it will open\na new one and any subsequent calls underneath that context manager will be added to the new transaction.\n\nIf `is_conditional` is `True` and there is no transaction in the call chain, then the connection will not create a new transaction. Instead, the query will be performed without a transaction.\n\n### Creating database and tables\n\nWith Pynocular you can use simple python code to create new databases and database tables. All you need is a working connection string to the database host, a `DatabaseInfo` object that contains the information of the database you want to create, and a properly decorated pydantic model. When you decorate a pydantic model with Pynocular, it creates a SQLAlchemy table as a private variable. This can be accessed via the `_table` property\n(although accessing private variables is not recommended).\n\n```python\nfrom pynocular.db_util import create_new_database, create_table\n\nfrom my_package import Org, db_info\n\nconnection_string = "postgresql://postgres:XXXX@localhost:5432/postgres?sslmode=disable"\n\n# Creates a new database and "organizations" table in that database\nawait create_new_database(connection_string, db_info)\nawait create_table(db_info, Org._table)\n\n```\n\n### Unit Testing with DatabaseModels\n\nPynocular comes with tooling to write unit tests against your DatabaseModels, giving you\nthe ability to test your business logic without the extra work and latency involved in\nmanaging a database. All you have to do is use the `patch_database_model` context\nmanager provided in pynocular.\n\n```python\nfrom pynocular.patch_models import patch_database_model\n\nfrom my_package import Org, User\n\n\nwith patch_database_model(Org):\n    orgs = [\n        Org(id=str(uuid4()), name="orgus borgus", slug="orgus_borgus"),\n        Org(id=str(uuid4()), name="orgus borgus2", slug="orgus_borgus"),\n    ]\n\n    await Org.create_list(orgs)\n    fetched_orgs = await Org.get_list(name=orgs[0].name)\n    assert orgs[0] == fetched_orgs[0]\n\n# patch_database_model also works with nested models\nusers = [\n    User(id=str(uuid4()), username="Bob"),\n    User(id=str(uuid4()), username="Sally"),\n]\norgs = [\n    Org(\n        id=str(uuid4()),\n        name="orgus borgus",\n        slug="orgus_borgus",\n        tech_owner=users[0],\n        business_owner=users[1],\n    ),\n]\n\nwith patch_database_model(Org, models=orgs), patch_database_model(\n    User, models=users\n):\n    org = await Org.get(orgs[0].id)\n    org.name = "new test name"\n    users[0].username = "bberkley"\n\n    # Save the username update when saving the org model update\n    await org.save(include_nested_models=True)\n\n    # Get the org with the resolved nested model\n    org_get = await Org.get_with_refs(org.id)\n    assert org_get.name == "new test name"\n    assert org_get.tech_owner.username == "bberkley"\n```\n\n## Development\n\nTo develop pynocular, install dependencies and enable the pre-commit hook:\n\n```bash\npip install pre-commit poetry\npoetry install\npre-commit install\n```\n\nTo run tests:\n\n```bash\npoetry run pytest\n```\n',
    'author': 'RJ Santana',
    'author_email': 'ssantana@narrativescience.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/NarrativeScience/pynocular',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.5,<4.0.0',
}


setup(**setup_kwargs)



"""Context manager for mocking db calls for DatabaseModels during tests"""
from contextlib import contextmanager
import functools
from typing import Any, Dict, List, Optional
from unittest.mock import patch
from uuid import uuid4

from sqlalchemy import Column
from sqlalchemy.sql.elements import (
    BinaryExpression,
    BindParameter,
    ClauseList,
    ColumnElement,
    False_,
    Grouping,
    Null,
    True_,
    UnaryExpression,
)
from sqlalchemy.sql.operators import in_op, is_

from pynocular.database_model import DatabaseModel


@contextmanager
def patch_database_model(
    model_cls: DatabaseModel,
    models: Optional[List[DatabaseModel]] = None,
) -> None:
    """Patch a DatabaseModel class, seeding with a set of values

    Example:
        with patch_database_model(Org, [Org(id="1", name="org 1"), ...]:
            await Org.get_list(...)

    Args:
        model_cls: A subclass of DatabaseModel that should be patched.
        models: models that should be in the patched DB table.

    """
    models = models or []

    def match(model: DatabaseModel, expression: BinaryExpression) -> bool:
        """Function to match the value with the expected one in the expression

        Args:
            model: The db model that represents a model in the "db".
            expression: The expression object to compare to.

        Returns:
            True if the expression operator is True.

        """
        return expression.operator(
            model.get(expression.left.name), expression.right.value
        )

    async def select(
        where_expressions: Optional[List[BinaryExpression]] = None,
        order_by: Optional[List[UnaryExpression]] = None,
        limit: Optional[int] = None,
    ) -> List[DatabaseModel]:
        """Mock select function for DatabaseModel

        Args:
            where_expressions: The BinaryExpressions to use in the select where clause.
            order_by: The order by expressions to be included in the select. This are
            not supported for mocking at this time.
            limit: The maximum number of objects to return.

        Returns:
            List of DatabaseModels that match the parameters.

        """
        # This function currently does not support `order_by` parameter.
        if where_expressions is None:
            return models

        matched_models = [
            model
            for model in models
            if all(
                _evaluate_column_element(expr, model.to_dict())
                for expr in where_expressions
            )
        ]

        if limit is None:
            matched_models[:limit]

        return matched_models

    async def create_list(models) -> List[DatabaseModel]:
        """Mock `create_list` function for DatabaseModel

        Args:
            models: List of DatabaseModels to persist.

        Returns:
            The list of new DatabaseModels that have been saved.

        """
        # Iterate through the list of orm objs and call save().
        for obj in models:
            await obj.save()

        return models

    async def save(model, include_nested_models=False) -> None:
        """Mock `save` function for DatabaseModel

        Args:
            model: The model to save.
            include_nested_models: If True, any nested models should get saved before
                this object gets saved.

        """
        # If include_neted_models is True, call save on all nested model attributes.
        # This requires that the nested models are also patched.
        if include_nested_models:
            for attr_name in model._nested_model_attributes:
                obj = getattr(model, attr_name)
                if obj is not None:
                    await obj.save()

        primary_keys = model_cls._primary_keys
        # Put uuids into any primary key that isn't set yet.
        for primary_key in primary_keys:
            val = getattr(model, primary_key.name)
            if val is None:
                setattr(model, primary_key.name, str(uuid4()))

        # Pull the primary keys out of the class and the values out of the provided
        # orm object. Then build a where_expression list to get the model matching those
        # primary keys.
        where_expressions = [
            primary_key == getattr(model, primary_key.name)
            for primary_key in primary_keys
        ]
        selected_models = [
            model
            for model in models
            if all(
                _evaluate_column_element(expr, model.to_dict())
                for expr in where_expressions
            )
        ]

        if len(selected_models) == 0:
            # Add a new model to the models since this model didn't exist before.
            models.append(model)
        else:
            # Update the matching model. Since these are primary keys there should only
            # ever be one model matching the given where_expressions.
            matched_model = selected_models[0]
            for attr, val in model.dict().items():
                setattr(matched_model, attr, val)

    async def update_record(**kwargs: str) -> DatabaseModel:
        """Mock `update_record` function for DatabaseModel

        Args:
            kwargs: The values to update.

        Returns:
            The updated DatabaseModel.

        """
        primary_keys = model_cls._primary_keys

        # Pull the primary keys out of the class and the values out of the provided
        # kwargs. Then build a where_expression list to get the model matching those
        # primary keys.
        where_expressions = [
            primary_key == kwargs[primary_key.name] for primary_key in primary_keys
        ]
        selected_models = [
            model
            for model in models
            if all(
                _evaluate_column_element(expr, model.to_dict())
                for expr in where_expressions
            )
        ]

        # Update the matching model. Since these are primary keys there should only
        # ever be one model matching the given where_expressions.
        model = selected_models[0]
        for attr, val in kwargs.items():
            setattr(model, attr, val)
        return model

    # Add the patches. Note that create functionality is patched indirectly though
    # 'save' already, but add a spy on it anyway so we can test calls against it.
    with patch.object(model_cls, "select", select), patch.object(
        model_cls, "save", save
    ), patch.object(model_cls, "update_record", update_record), patch.object(
        model_cls, "create_list", create_list
    ), patch.object(
        model_cls, "create", wraps=model_cls.create
    ):
        yield


@functools.singledispatch
def _evaluate_column_element(
    column_element: ColumnElement, model: Dict[str, Any]
) -> Any:
    """Evaluate a ColumnElement on a dictionary representing a database model

    This function can be overridden based on the type of ColumnElement to return
    an element from the model, a static value, or the result of some operation (e.g.
    addition).

    Args:
        column_element: The element to evaluate.
        model: The model to evaluate the column element on. Represented as a dictionary
            where the keys are column names.

    """
    raise Exception(f"Cannot evaluate a {column_element} object.")


@_evaluate_column_element.register(ClauseList)
def _evaluate_clause_list(column_element: ClauseList, model: Dict[str, Any]) -> Any:
    """Evaluates a clause list and breaks it down into its sub column elements

    Args:
        column_element: The ClauseList in question.
        model: The model of data this clause should be evaluated for.

    Returns:
        The result of the evaluation.

    """
    operator = column_element.operator

    return operator(
        *[
            _evaluate_column_element(sub_element, model)
            for sub_element in column_element.get_children()
        ]
    )


@_evaluate_column_element.register(BinaryExpression)
def _evaluate_binary_expression(
    column_element: BinaryExpression, model: Dict[str, Any]
) -> Any:
    """Evaluates the binary expression

    Args:
        column_element: The binary expression to evaluate.
        model: The model to evaluate the expression on.

    Returns:
        The evaluation response dictated by the operator of the expression.

    """
    operator = column_element.operator

    # The sqlalchemy `in` operator does not work on evaluated columns, so we replace
    # it with the standard `in` operator.
    if operator == in_op:
        operator = lambda x, y: x in y

    # The sqlalchemy `is` operator does not work on evaluated columns, so we replace it
    # with the standard `is` operator.
    if operator == is_:
        operator = lambda x, y: x is y

    return operator(
        _evaluate_column_element(column_element.left, model),
        _evaluate_column_element(column_element.right, model),
    )


@_evaluate_column_element.register(Column)
def _evaluate_column(column_element: Column, model: Dict[str, Any]) -> Any:
    """Evaluate the column based on the column element name

    Args:
        column_element: The column to evaluate.
        model: The model dictionary.

    Returns:
        The value from the model of attribute referenced by column_element.

    """
    return model.get(column_element.name)


@_evaluate_column_element.register(BindParameter)
def _evaluate_bind_parameter(
    column_element: BindParameter, model: Dict[str, Any]
) -> Any:
    """Evaluate the column_elements value

    Args:
        column_element: The column's bind parameter.
        model: The model dictionary.

    Returns:
        The value of the column_element

    """
    return column_element.value


@_evaluate_column_element.register(True_)
def _evaluate_true(column_element: True_, model: Dict[str, Any]) -> bool:
    """Wrapper around evaluating True

    Args:
        column_element: The column to evaluate. This is just True
        model: The model dictionary.

    Returns:
        True

    """
    # The boolean value True is its own SQLAlchemy element
    return True


@_evaluate_column_element.register(False_)
def _evaluate_false(column_element: False_, model: Dict[str, Any]) -> bool:
    """Wrapper around evaluating False

    Args:
        column_element: The column to evaluate. This just returns False
        model: The model dictionary.

    Returns:
        False

    """
    # The boolean value False is its own SQLAlchemy element
    return False


@_evaluate_column_element.register(Grouping)
def _evaluate_grouping(column_element: Grouping, model: Dict[str, Any]) -> List[Any]:
    """Wrapper around evaluating a grouping

    Args:
        column_element: The grouping to evaluate.
        model: The model dictionary.

    Returns:
        A list of of values that are the resulting of evaluating each element in the
        group.

    """
    return [
        _evaluate_column_element(clause, model)
        for clause in column_element.element.clauses
    ]


@_evaluate_column_element.register(Null)
def _evaluate_null(column_element: Null, model: Dict[str, Any]) -> Any:
    """Wrapper around evaluating null

    Args:
        column_element: The column element to evaluate. This is null
        model: The model dictionary.

    Returns:
        None

    """
    # The Null value is its own SQLAlchemy element
    return None
